// ---- jobs2.c ----

#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>

#include "syntax.h"
#include "jobs.h"

#define N_SIGS 9

job_t job_tab[MAX_JOBS];
int first_free = 1;

typedef struct st_sig_name_t {
  int num;
  char *name;
} sig_name_t;

sig_name_t sig_name_tab [N_SIGS] =
{ { SIGTERM, "SIGTERM" },
  { SIGKILL, "SIGKILL" },
  { SIGCONT, "SIGCONT" },
  { SIGSTOP, "SIGSTOP" },
  { SIGTSTP, "SIGTSTP" },
  { SIGHUP,  "SIGHUP" },
  { SIGSEGV, "SIGSEGV" },
  { SIGUSR1, "SIGUSR1" },
  { SIGUSR2, "SIGUSR2" } };

char *signame (int signum) {
  int i;
  static char def[20];
  for (i = 0; i < N_SIGS; i++)
    if (sig_name_tab[i].num == signum)
      return sig_name_tab[i].name;

  sprintf (def, "SIG %d", signum);
  return def;
}

void init_job_tab (void)
{
  int id;
  for (id = 0; id < MAX_JOBS; id++)
    job_tab[id] = NULL;
  first_free = 1;
}

int can_create_new_job (void)
{
  return first_free < MAX_JOBS;
}

int exists_job (void)
{
  return first_free > 1;
}

job_t make_job (pipe_t pipe)
{
  job_t l, *p;

  p = &l;
  while (pipe) {
    *p = malloc (sizeof **p);
    (*p)->name = calloc (1 + strlen (pipe->scomm.name), 
			 sizeof (char));
    strcpy ((*p)->name, pipe->scomm.name);
    (*p)->pid = pipe->pid;
    (*p)->status = RUN;
    (*p)->code = 0;
    p = &((*p)->next);
    pipe = pipe->next;
  }
  *p = NULL;

  return l;
}

void print_summary_job (FILE *out, int id)
{
  job_t l;
  
  l = job_tab[id];
  fprintf (out, "[%d]", id);
  while (l) {
    fprintf (out, " %d", (int) l->pid);
    l = l->next;
  }
  fprintf (out, "\n");
}

int add_job (FILE *out, job_t job)
{
  if (!can_create_new_job())
    return -1;

  if (!job)
    return 0;

  job_tab[first_free] = job;
  print_summary_job (out, first_free);

  return first_free++;
}


int remove_job (int id)
{
  job_t l, p;

  if (id < 0 || first_free <= id || !job_tab[id])
    return -1;

  l = job_tab[id];
  while (l) {
    p = l;
    l = l->next;
    free (p->name);
    free (p);
  }

  job_tab[id] = NULL;
  if (id == first_free - 1) {
    do first_free--;
    while (first_free > 0 && job_tab[first_free] == NULL);
    first_free++;
  }

  return 0;
}


int convert_status (int status, job_t l)
{
  if (WIFEXITED (status)) {
    l->status = EXIT;
    l->code = WEXITSTATUS (status);
    return 0;
  }
  if (WIFSTOPPED (status)) {
    l->status = STOP;
    l->code = WSTOPSIG (status);
    return 1;
  }
  if (WIFSIGNALED (status)) {
    l->status = SIG;
    l->code = WTERMSIG (status);
    return 0;
  }
  return 1;
}


int update_proc_status (job_t l)
{
  int st, r;

  if (l->status == EXIT || l->status == SIG)
    return 0;

  if ((r = waitpid (l->pid, &st, WNOHANG | WUNTRACED)) < 0) {
    return 0;
  }
  
  if (r == 0) {
    l->status = RUN;
    l->code = 0;
    return 1;
  }
  
  return convert_status (st, l);
  
}

int wait_job (int id, int *status)
{
  job_t l;
  int r;

  if (id < 0 || MAX_JOBS <= id || !job_tab[id])
    return -1;
  
  l = job_tab[id];
  while (l) {
    if (l->pid != (pid_t) -1 &&
	(l->status == RUN || l->status == STOP)) {
      if ((r = waitpid (l->pid, status, 0)) < 0) {
	perror ("waitpid in wait_job");
	return -1;
      } else if (r >= 0)
	convert_status (*status, l);
    }
    l = l->next;
  }
  remove_job (id);
  return 0;
}

int put_in_fg_job (job_t job, int *status)
{
  if (job_tab[0])
    return -1;

  job_tab[0] = job;
  wait_job (0, status);
  return 0;
}

int put_fg_in_bg_job (void)
{
  int r;

  r = add_job (stdout, job_tab[0]);
  if (r > 0)
    job_tab[0] = NULL;
  return r;
}

int kill_job (int id, int sig)
{
  job_t l;
  int alive_procs = 0, all_stopped = 1;

  if (id < 0 || MAX_JOBS <= id || !job_tab[id])
    return -1;

  l = job_tab[id];
  while (l) {
    if (update_proc_status (l)) {
      if (kill (l->pid, sig) <0) {
	return -1;
      }
      if (update_proc_status (l))
	alive_procs = 1;
    }
    if (l->status == RUN)
      all_stopped = 0;
    l = l->next;
  }
  if (!alive_procs)
    return remove_job (id);

  if (id == 0 && all_stopped)
    return put_fg_in_bg_job ();

  return 0;
}

int kill_jobs (int sig)
{
  int id;
  
  for (id = 0; id < first_free; id++)
    if (job_tab[id]) {
      kill_job (id, sig);
      fprintf (stderr, "Warning: %s sent to job %d\n", signame (sig), id);
    }

  return 0;
}

void print_proc (FILE *out, job_t l)
     /* affiche l'tat du processus point par l
	(ne parcourt pas l!) */
{
  fprintf (out, "\t%d\t", (int) l->pid);
  switch (l->status) {
  case RUN:  fprintf (out, "Running       "); 
    break;
  case STOP: fprintf (out, "Stopped");
    switch (l->code) {
    case SIGTSTP: fprintf (out, " by tty");
      break;
    case SIGSTOP: fprintf (out, "       ");
      break;
    default: fprintf (out, " by %2d ", l->code);
    }
    break;
  case EXIT: 
    if (l->code)
      fprintf (out, "Exited %3d    ", l->code);
    else 
      fprintf (out, "Done          ");
    break;
  case SIG:
    switch (l->code) {
    case SIGTERM: fprintf (out, "Terminated    ");
      break;
    case SIGKILL: fprintf (out, "Killed        ");
      break;
    case SIGSEGV: fprintf (out, "Seg. fault    ");
      break;
    case SIGHUP: fprintf (out, "Hanguped      ");
      break;
    case SIGPIPE: fprintf (out, "Broken pipe   ");
      break;
    default: fprintf (out, "%sed     ", signame (l->code));
    }
    break;
  default: fprintf (out, "Unknown status");
  }
  if (l->next)
    fprintf (out, "%s |\n", l->name);
  else 
    fprintf (out, "%s &\n", l->name);
  
}

void update_print_job (FILE *out, int id)
     /* actualise et affiche les informations
	du job de n id */
{
  job_t l;
  int alive_procs = 0;

  if (!job_tab[id])
    return;

  l = job_tab[id];
  fprintf (out, "[%d]", id);
  while (l) {
    if (update_proc_status (l))
      alive_procs = 1;
    print_proc (out, l);
    l = l->next;
  }
  if (!alive_procs)
    remove_job (id);
}

void update_print_jobs (FILE *out)
{
  int id;

  for (id = 1; id < first_free; id++)
    update_print_job (out, id);
}

void update_print_if_dead_job (FILE *out, int id)
     /* actualise et ( affiche les informations
	du job de n id si il est vivant ) */
{
  job_t l;
  int alive_procs = 0;

  if (!job_tab[id])
    return;

  l = job_tab[id];
  while (l) {
    if (update_proc_status (l))
      alive_procs = 1;
    l = l->next;
  }
  
  if (!alive_procs) {
    l = job_tab[id];
    fprintf (out, "[%d]", id);
    while (l) {
      print_proc (out, l);
      l = l->next;
    }
    remove_job (id);
  }
}

void update_print_if_dead_jobs (FILE *out)
{
  int id;

  for (id = 0; id < first_free; id++) 
    update_print_if_dead_job (out, id);
  
}


